import json
from pytest import mark

from django.test import TestCase, Client

from channels.db import database_sync_to_async as db
from reactor.tests import reactor

from .models import Item


class TestNormalRendering(TestCase):

    def setUp(self):
        Item.objects.create(text='First task')
        Item.objects.create(text='Second task')
        self.c = Client()

    def test_everything_two_tasks_are_rendered(self):
        response = self.c.get('/')
        assert response.status_code == 200
        self.assertContains(response, 'First task')
        self.assertContains(response, 'Second task')


@mark.asyncio
@mark.django_db(transaction=True)
async def test_live_components():
    client = Client()
    assert await db(Item.objects.count)() == 0

    html_response = await db(client.get)('/')
    assert html_response.status_code == 200

    async with reactor() as comm:

        x_list = comm.add_component(
            'x-todo-list',
            {'id': 'someid', 'showing': 'all'}
        )
        x_todo_counter = comm.add_component(
            'x-todo-counter',
            {'id': 'someid-counter'}
        )
        doc = await comm.send_join(x_list)
        todo_list = doc('#someid')
        assert json.loads(todo_list.attr['state']) == comm[x_list].state

        # Add new item
        doc = await comm.send(x_list, 'add', new_item='First task')

        # There was an item crated and rendered
        assert await db(Item.objects.count)() == 1
        assert len(doc('[is=x-todo-item]')) == 1
        todo_item_id = doc('[is=x-todo-item]')[0].get('id')
        todo_item_label = doc('[is=x-todo-item] label')[0]
        assert todo_item_label.text == 'First task'

        item = await db(Item.objects.first)()
        assert str(item.id) == todo_item_id
        assert item.text == 'First task'

        # Click title to edit it
        assert len(doc('[is=x-todo-item] li.editing')) == 0
        x_first_item = comm.add_component('x-todo-item', {'id': todo_item_id})
        doc = await comm.send(x_first_item, 'toggle_editing')
        assert len(doc('[is=x-todo-item] li.editing')) == 1

        # Edit item with a new text
        doc = await comm.send(x_first_item, 'save', text='Edited task')
        assert len(doc('[is=x-todo-item] li.editing')) == 0, 'Not in read mode'
        todo_item_label = doc('[is=x-todo-item] label')[0]
        assert todo_item_label.text == 'Edited task'
        await db(item.refresh_from_db)()
        assert item.text == 'Edited task'

        # Check counter has 1 item left
        doc = comm[x_todo_counter].doc
        assert doc('strong')[0].text == '1'

        # Mark item as completed
        assert len(doc('[is=x-todo-item] li.completed')) == 0
        doc = await comm.send(x_first_item, 'completed', completed=True)

        # Item is completed
        assert len(doc('[is=x-todo-item] li.completed')) == 1

        # Counter is rendered as 0
        doc = comm[x_todo_counter].doc
        assert doc('strong')[0].text == '0'

        # Switch to "only active tasks" filtering
        doc = await comm.send(x_list, 'show', showing='active')
        assert len(doc('[is=x-todo-item] li.hidden')) == 1

        # Switch to "only done tasks" filtering
        doc = await comm.send(x_list, 'show', showing='completed')
        assert len(doc('[is=x-todo-item]')) == 1
        assert len(doc('[is=x-todo-item] li.hidden')) == 0

        # Switch to "all tasks" filtering
        doc = await comm.send(x_list, 'show', showing='all')
        assert len(doc('[is=x-todo-item]')) == 1
        assert len(doc('[is=x-todo-item] li.hidden')) == 0

        # Add another task
        doc = await comm.send(x_list, 'add', new_item='Another task')
        assert len(doc('[is=x-todo-item]')) == 2
        assert len(doc('[is=x-todo-item] li.completed')) == 1

        # Cache miss in the counter so re-render
        doc = comm[x_todo_counter].doc
        assert doc('strong')[0].text == '1'

        # Remove completed tasks removes just one task
        doc = await comm.send(x_list, 'clear_completed')
        assert len(doc('[is=x-todo-item]')) == 1
        assert len(doc('[is=x-todo-item].completed')) == 0

        # Cache miss in the counter so re-render
        doc = comm[x_todo_counter].doc
        assert doc('strong')[0].text == '1'



from typing import Generator

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models


class Encoder(DjangoJSONEncoder):

    def default(self, o):
        if isinstance(o, models.Model):
            return o.pk

        if isinstance(o, models.QuerySet):
            return list(o.values_list('pk', flat=True))

        if isinstance(o, (Generator, set)):
            return list(o)

        if hasattr(o, '__json__'):
            return o.__json__()

        return super().default(o)



from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter

from reactor.urls import websocket_urlpatterns

application = ProtocolTypeRouter({
    'websocket': AuthMiddlewareStack(URLRouter(
        websocket_urlpatterns,
    ))
})



from uuid import uuid4
from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from reactor import broadcast


class BaseModel(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4, editable=False)

    class Meta:
        abstract = True


class ItemQS(models.QuerySet):

    @property
    def completed(self):
        return self.filter(completed=True)

    @property
    def active(self):
        return self.filter(completed=False)

    def update(self, *args, **kwargs):
        results = super().update(*args, **kwargs)
        broadcast('item')
        broadcast('item.updated')
        for item in self:
            broadcast(f'item.{item.id}')
        return results


class Item(BaseModel):
    completed = models.BooleanField(default=False)
    text = models.CharField(max_length=256)

    objects = ItemQS.as_manager()

    def __str__(self):
        return self.text


@receiver(post_save, sender=Item)
def emit_element_saved(sender, instance, created, **kwargs):
    broadcast(f'item')
    if created:
        broadcast('item.new')
    else:
        broadcast('item.updated')
        broadcast(f'item.{instance.id}')


@receiver(post_delete, sender=Item)
def emit_element_deleted(sender, instance, **kwargs):
    broadcast('item')
    broadcast('item.deleted')
    broadcast(f'item.{instance.id}')



# Generated by Django 2.2 on 2019-04-17 21:27

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('completed', models.BooleanField(default=False)),
                ('text', models.CharField(max_length=256)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.db.transaction import atomic
from django.shortcuts import render
from reactor.component import Component

from .models import Item


def todo(request):
    return render(request, 'todo.html')


class XTodoList(Component):
    template_name = 'todo/list.html'

    def mount(self, showing='all', **kwargs):
        self.showing = showing
        self.subscribe('item.new')

    def serialize(self):
        return dict(
            id=self.id,
            showing=self.showing,
        )

    @property
    def items(self):
        return Item.objects.all()

    @property
    def all_items_are_completed(self):
        return self.items.count() == self.items.completed.count()

    @atomic
    def receive_add(self, new_item, **kwargs):
        Item.objects.create(text=new_item)

    def receive_show(self, showing, **kwargs):
        self.showing = showing

    @atomic
    def receive_toggle_all(self, toggle_all, **kwargs):
        self.items.update(completed=toggle_all)

    @atomic
    def receive_clear_completed(self, **kwargs):
        self.items.completed.delete()


class XTodoCounter(Component):
    template_name = 'todo/counter.html'

    def mount(self, items=None, *args, **kwargs):
        self.items = items or Item.objects.all()
        self.subscribe('item')


class XTodoItem(Component):
    template_name = 'todo/item.html'

    def mount(self, item=None, editing=False, showing='all', **kwargs):
        self.editing = editing
        self.showing = showing
        self.item = item or Item.objects.filter(id=self.id).first()
        if self.item:
            self.subscribe(f'item.{self.item.id}')
        else:
            self.send_destroy()

    def serialize(self):
        return dict(
            id=self.id,
            editing=self.editing,
            showing=self.showing,
        )

    def is_visible(self):
        return (
            self.showing == 'all' or
            self.showing == 'completed' and self.item.completed or
            self.showing == 'active' and not self.item.completed
        )

    @atomic
    def receive_destroy(self, **kwargs):
        self.item.delete()

    @atomic
    def receive_completed(self, completed, **kwargs):
        self.item.completed = completed
        self.item.save()

    def receive_toggle_editing(self, **kwargs):
        if not self.item.completed:
            self.editing = not self.editing

    @atomic
    def receive_save(self, text, **kwargs):
        self.item.text = text
        self.item.save()
        self.editing = False


class XCounter(Component):

    template_name = 'x-counter.html'

    def mount(self, amount=0, **kwargs):
        self.amount = amount

    def serialize(self):
        return dict(id=self.id, amount=self.amount)

    def receive_inc(self, **kwargs):
        self.amount += 1

    def receive_dec(self, **kwargs):
        self.amount -= 1

    def receive_set_to(self, amount, **kwargs):
        self.amount = amount


def counter(request):
    return render(request, 'counter.html')



"""
Django settings for fision project.

Generated by 'django-admin startproject' using Django 2.2.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os
up = os.path.dirname

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = up(up(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'b7!j@8gk-vdq3tona^(i(qg#xiir*%r-@u1f&fw@@(ccwy^ijb'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'fision.todo',
    'reactor',

    'channels',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'fision.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'fision.wsgi.application'
ASGI_APPLICATION = 'fision.asgi.application'

# In memory
CHANNEL_LAYERS = {
    'default': {
        'BACKEND': 'channels.layers.InMemoryChannelLayer',
    },
}

# # With redis
# CHANNEL_LAYERS = {
#     'default': {
#         'BACKEND': 'channels_redis.core.RedisChannelLayer',
#         'CONFIG': {
#             "hosts": [('127.0.0.1', 6379)],
#         },
#     },
# }

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        'ATOMIC_REQUESTS': True,
        'OPTIONS': {
            'timeout': 20,
        },
        "TEST": {
            "NAME": os.path.join(BASE_DIR, "db_test.sqlite3"),
        },
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',  # noqa
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',  # noqa
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',  # noqa
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',  # noqa
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'django.server',
        },
    },
    'formatters': {
        'django.server': {
            '()': 'reactor.log.ServerFormatter',
            'format': '[{server_time}] {message}',
            'style': '{',
        }
    },
    'loggers': {
        'reactor': {
            'handlers': ['console'],
            'level': 'DEBUG',

        },
    },
}
